% function [ok, li, cp, tc] = trace_ray( angle, op, grid, gcs, ng )
% input:
%   angle .. obsetver view vector angle
%   op .. observer point
%   grid .. maze grid
%   gcs .. grid cell sizes
%   ng .. grid sizes
% output:
%   ok .. something was hit?
%   li .. line intersection point
%   cp .. hit cell coordinates
%   tc .. texture column corresponding to li
function [ok, li, cp, tc] = trace_ray( angle, op, grid, gcs, ng )
global g_zero

% Observer angle, vector.
oa = angle * pi / 180;
ov = [cos(oa), -sin(oa); sin(oa), cos(oa)] * [1; 0];

% Observer point in grid coordinates.
opg = floor( op ./ gcs );

% Check point vector.
cpv = [-(ov(1) < -g_zero); -(ov(2) < -g_zero)];

% X-line intersections.
if (ov(1) > g_zero)
  i = opg(1) + 1;
  ok_x = 0;
  while (i < ng(1))
    s = (i * gcs(1) - op(1)) / ov(1);
    xli = round( op + s * ov );
    % Checkpoint.
    xcp = floor( (xli + cpv) ./ gcs ) + 1;
    if (any( [xcp(1) <= 0, xcp(2) <= 0, xcp(1) > ng(1), xcp(2) > ng(2)] ))
      break;
    end
    if (grid(xcp(2),xcp(1)) ~= 0)
      ok_x = 1;
      break
    end
    i = i + 1;
  end
%  [ok, xli, cp, cell] = intersect_x_plus( opg, op, ov, gcs, ng, grid );
elseif (ov(1) < -g_zero)
  i = opg(1);
  ok_x = 0;
  while (i > 0)
    s = (i * gcs(1) - op(1)) / ov(1);
    xli = round( op + s * ov );
    % Checkpoint.
    xcp = floor( (xli + cpv) ./ gcs ) + 1;
    if (any( [xcp(1) <= 0, xcp(2) <= 0, xcp(1) > ng(1), xcp(2) > ng(2)] ))
      break;
    end
    if (grid(xcp(2),xcp(1)) ~= 0)
      ok_x = 1;
      break
    end
    i = i - 1;
  end
else
  ok_x = 0;
end
% Y-line intersections.
if (ov(2) > g_zero)
  i = opg(2) + 1;
  ok_y = 0;
  while (i < ng(2))
    s = (i * gcs(2) - op(2)) / ov(2);
    yli = round( op + s * ov );
    % Checkpoint.
    ycp = floor( (yli + cpv) ./ gcs ) + 1;
    if (any( [ycp(1) <= 0, ycp(2) <= 0, ycp(1) > ng(1), ycp(2) > ng(2)] ))
      break;
    end
    if (grid(ycp(2),ycp(1)) ~= 0)
      ok_y = 1;
      break
    end
    i = i + 1;
  end
elseif (ov(2) < -g_zero)
  i = opg(2);
  ok_y = 0;
  while (i > 0)
    s = (i * gcs(2) - op(2)) / ov(2);
    yli = round( op + s * ov );
    % Checkpoint.
    ycp = floor( (yli + cpv) ./ gcs ) + 1;
    if (any( [ycp(1) <= 0, ycp(2) <= 0, ycp(1) > ng(1), ycp(2) > ng(2)] ))
      break;
    end
    if (grid(ycp(2),ycp(1)) ~= 0)
      ok_y = 1;
      break
    end
    i = i - 1;
  end
else
  ok_y = 0;
end

ok = 1;
tor = 0;
if (ok_x & ok_y)
  % Choose the nearer intersection.
  xdist = norm( xli - op );
  ydist = norm( yli - op );
  if (xdist < ydist)
    li = xli;
    cp = xcp;
    tc = li(2) - (cp(2) - 1) * gcs(2);
    tor = (ov(1) > g_zero);
  else
    li = yli;
    cp = ycp;
    tc = li(1) - (cp(1) - 1) * gcs(1);
    tor = (ov(2) < -g_zero);
  end
elseif (ok_x)
  li = xli;
  cp = xcp;
  tc = li(2) - (cp(2) - 1) * gcs(2);
  tor = (ov(1) > g_zero);
elseif (ok_y)
  li = yli;
  cp = ycp;
  tc = li(1) - (cp(1) - 1) * gcs(1);
  tor = (ov(2) < -g_zero);
else
  ok = 0;
  li = [0; 0];
  cp = [0; 0];
  tc = 0;
end
% Fix zero tc.
tc = (ok & ~tc) + tc;
% Fix texture orientation.
tc = tor * (gcs(1) - tc + 1) + (1 - tor) * tc;
